.nautilus-circular-button {
  border-radius: 20px;
  -gtk-outline-radius: 20px;
}

.nautilus-desktop-window,
.nautilus-desktop-window paned,
.nautilus-desktop-window notebook,
.nautilus-desktop-window notebook > stack {
    background: transparent;
}

.nautilus-menu-sort-heading {
    min-height: 26px;
    padding-left: 5px;
    padding-right: 5px;
}
