/*
   Default file icon used by the icon factory.
 
   Copyright (C) 2000 Eazel, Inc.
  
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
  
   You should have received a copy of the GNU General Public
   License along with this program; if not, see <http://www.gnu.org/licenses/>.
  
   Author: Darin Adler <darin@bentspoon.com>
*/

#ifndef NAUTILUS_DEFAULT_FILE_ICON_H
#define NAUTILUS_DEFAULT_FILE_ICON_H

extern const int nautilus_default_file_icon_width;
extern const int nautilus_default_file_icon_height;
extern const unsigned char nautilus_default_file_icon[];

#endif /* NAUTILUS_DEFAULT_FILE_ICON_H */
